structure Unparser  = 
struct 
open DataTypes;
fun indent n = 
    let fun mk_str size = if size < 1 then "" else " " ^ mk_str(size - 1);
    in
      print(mk_str(n))
    end

fun print_indent(str, margin) = (indent(margin);
                                 print(str))

fun say(Program(id(name), members), margin) =
     (print("class " ^ name ^ " { \n"); 
      say_members(members, margin + 1);
      print("\n}\n"))
and 
    say_members(MemberDecls(fields, methods), margin) =
     (say_decls(fields, margin); 
      print("\n");
      say_decls(methods, margin))
and 
     say_unary_op(NOT) = print("!")
   | say_unary_op(Cast(tipe)) = (print("("); say_type(tipe); print(")"))
and 
     say_exprs(exprs, margin) = List.map (fn e => say_expr(e)) exprs
and 
      say_expr(NoneExpr) = ()
    | say_expr(ArithmeticBinaryOperation(lhs, arit_binary_operator, rhs, _, _)) =
    (say_expr(lhs);
     say_arit_binary_op(arit_binary_operator);
     say_expr(rhs))
    | say_expr(LogicalBinaryOperation(lhs, logi_binary_operator, rhs, _, _)) =
    (say_expr(lhs);
     say_logi_binary_op(logi_binary_operator);
     say_expr(rhs))
    | say_expr(RelationalBinaryOperation(lhs, rel_binary_operator, rhs, _, _)) =
    (say_expr(lhs);
     say_rela_binary_op(rel_binary_operator);
     say_expr(rhs))
    | say_expr(EqualityBinaryOperation(lhs, equal_binary_operator, rhs, _, _)) =
    (say_expr(lhs);
     say_equal_binary_op(equal_binary_operator);
     say_expr(rhs))
    | say_expr(UnaryOperation (unary_op, expr, _, _)) = (say_unary_op(unary_op); say_expr(expr))
    | say_expr(CharLit(the_char, _, _)) = (print the_char)
    | say_expr(IntLit(the_num, _, _)) = (print(Int32.toString(the_num)))
    | say_expr(StrLit(the_str, _, _)) = (print the_str)
    | say_expr(True(_)) = (print "true")
    | say_expr(False(_)) = (print "false")
    | say_expr(ArglessFnCall(id(name), _, _)) = 
        (print name; 
         print "()")
    | say_expr(FnCall(id(name), actual_args, _, _)) = 
        (print name; 
         print "(";
         say_actual_args(actual_args);  
         print ")")
    | say_expr(Name(id(name), _, _)) = (print(name))
    | say_expr(ArrayDeref(id(name), expr, _, _)) = (print(name^"["); say_expr(expr); print "]")
    | say_expr(ActualArgument(expr, _, _)) = say_expr(expr)
    | say_expr(Checked(_, _)) = () (* We won't get these during the unparse. *)
and 
     say_actual_args(args) = List.map (fn e =>(say_expr(e); print(", "))) args
and 
    say_decls(decls, margin) = List.map (fn d => say_decl(d, margin)) decls
and
    say_decl(Field(Array(tipe, size), id(name), _, _), margin) = 
        (indent(margin);
         say_type(tipe);
         print (name^"["^Int32.toString(size)^"];\n"))
    | say_decl(Field(tipe, id(s), _, _), margin) = 
        (indent(margin);
         say_type(tipe);
         print s;
         print ";\n")
    | say_decl(InitializedField(UnknownConst, id(s), value, _, _), margin) =
        (indent(margin);
         print "const ";
         print (s^" = ");
         say_expr(value);
         print ";\n")
    | say_decl(InitializedField(tipe, id(s), value, _, _), margin) =
        (indent(margin);
         say_type(tipe);
         print (s^" = ");
         say_expr(value);
         print ";\n")
    | say_decl(FormalArg(ArrayParam(tipe), id(name), _, _), margin) = (say_type(tipe); print(name^"[], "))
    | say_decl(FormalArg(tipe, id(name), _, _), margin) = (say_type(tipe); print(name^", "))
    | say_decl(MainMethod(field_decls, stmts, _, _), margin) = 
        say_decl(Method(Void, id("main"), [], field_decls, stmts, 0, 0), margin)
    | say_decl(Method(tipe, id(name), arg_decls, field_decls, stmts, _, _), margin) = 
      (indent(margin);
       say_type(tipe);
       print(name^"(");
       say_decls(arg_decls, margin);
       print(") { \n");
       say_decls(field_decls, margin + 1);
       say_stmts(stmts, margin + 1);
       indent(margin);
       print("}\n"))
     | say_decl(NoneDecl, margin) = () 
and 
     say_stmts(stmts, margin) = List.map (fn stmt => say_stmt(stmt, margin)) stmts
and 
       say_stmt(NoneStmt, margin) = ()
     | say_stmt(Block(decls, stmts, _, _), margin) = 
          (indent(margin); 
           print "{";
           say_decls(decls, margin + 1);
           say_stmts(stmts, margin + 1);
           indent(margin);
           print "}\n")
     | say_stmt(IfThen(cond, then_stmt, _, _), margin) = 
          (print_indent("if(", margin);
           say_expr(cond);
           print(") { \n");
           say_stmt(then_stmt, margin + 1);
           print ("\n");
           print_indent("}\n", margin))
     | say_stmt(IfThenElse(cond, then_stmt, else_stmt, _, _), margin) = 
           (say_stmt(IfThen(cond, then_stmt, 0, 0), margin);
            print " else {\n";
            say_stmt(else_stmt, margin + 1);
            print "\n";
            print_indent("}\n", margin))
     | say_stmt(While(cond, stmt, _, _), margin) = 
            (print_indent("while( ", margin);
             say_expr(cond);
             print ")\n";
             say_stmt(stmt, margin + 1))
     | say_stmt(LabeledWhile(Label(id(name)), cond, stmt, _, _), margin) = 
            (print_indent(name^": ", margin);
             say_stmt(stmt, margin))
     | say_stmt(LabeledWhile(_, cond, stmt, _, _), margin) = (print("FIXME: Labels(n)"))
     | say_stmt(Asg(lval, rval, _, _), margin) = (
        indent(margin);
        say_expr(lval);
        print " = "; 
        say_expr(rval);
        print ";\n")
     | say_stmt(Read(expr, _, _), margin) = (say_expr(expr))
     | say_stmt(Reads(reads, _, _), margin) = (
        print_indent("read(", margin);
        List.map (fn r => (say_stmt(r, margin); print ", "))  reads;
        print ");\n"
        )
     | say_stmt(Writes(writes, _, _), margin) =  (
        print_indent("print(", margin);
        List.map (fn w => (say_stmt(w, margin); print ", "))  writes;
        print ");\n"
        )
     | say_stmt(Write(expr, _, _), margin) = (say_expr(expr))
     | say_stmt(ArglessProcCall(id(name), _, _), margin) = (print_indent(name^"();\n", margin))
     | say_stmt(ProcCall(id(name), args, _, _), margin) = (print_indent(name^"();\n", margin))
     | say_stmt(Return(NoneExpr, _, _), margin) = (print_indent("return;", margin))
     | say_stmt(Return(expr, _, _), margin) = (print_indent("return ", margin); say_expr(expr); print ";\n")
     | say_stmt(Break(Label(id(name)), _, _), margin) = (print_indent(("break "^name^";"), margin))
     | say_stmt(Break(_, _, _), margin) = ()
     | say_stmt(Continue(Label(id(name)), _, _), margin) = ()
     | say_stmt(Continue(_, _, _), margin) = ()
     | say_stmt(Exit(code, _, _), margin) = (print_indent("exit(", margin);
     print(Int32.toString(code)); print ");\n")

and    say_type(Int)  = (print "int ")
     | say_type(Char) = (print "char ")
     | say_type(Bool) = (print "bool ")
     | say_type(Void) = (print "void ")
     | say_type(UnknownConst) = (print "const ")
     | say_type(Const(x)) = (say_type(x))
     | say_type(Param(x)) = (say_type(x))
     | say_type(Array(x, n)) = (say_type(x))
     | say_type(ArrayParam(x)) = (say_type(x))
     | say_type(StringLit(n)) = ()
     | say_type(NoneType) = ()
     | say_type(Error) = ()
and
    say_rela_binary_op(LT) = print "<"
  | say_rela_binary_op(GT) = print ">"
  | say_rela_binary_op(LEQ) = print "<="
  | say_rela_binary_op(GEQ) = print ">="
and
    say_equal_binary_op(EQ) = print "=="
  | say_equal_binary_op(NOTEQ) = print "!="
and
    say_arit_binary_op(PLUS) = print "+"
  | say_arit_binary_op(MINUS) = print "-"
  | say_arit_binary_op(TIMES) = print "*"
  | say_arit_binary_op(DIV) = print "/"
and
   say_logi_binary_op(AND) = print "&&"
 | say_logi_binary_op(OR) = print "||"

fun unparse filename =
    let val (tree, _) = Parser.parse(filename)
    in    
      say(tree, 0); 
      tree
    end

fun unparse_all _ =
  List.map (fn testid => 
       let val testid_str = if testid < 10 then
                          "0"^(Int.toString testid)
                        else
                          Int.toString testid 
        in
          unparse("tests/p"^testid_str^"csx.csx")
       end)
      (List.tabulate (69, (fn x => x))); (*FIXME 70*)

fun test testid = unparse("tests/p"^testid^"csx.csx");
end;
